from typing import List, Optional, Tuple


class BaseStorage:
    def path_join(self, path, *segments):
        raise NotImplementedError  # pragma: no cover

    def file_exists(self, file: List[str]) -> bool:
        raise NotImplementedError  # pragma: no cover

    def read_text(self, file: List[str], encoding: str = 'utf-8') -> str:
        raise NotImplementedError  # pragma: no cover

    def batch_change_files(self, changes: List[Tuple[Optional[str], List[str]]]):
        raise NotImplementedError  # pragma: no cover



import io
import warnings
from operator import itemgetter
from typing import Tuple, List, Mapping, Any, Iterable, Union, Dict, Callable

from hbutils.string.tree import format_tree

from .item import create_sync_item, SyncItem, _PRESERVED_NAMES, _REGISTERED_SYNC_TYPES
from ..utils import hash_anything, to_segments, TargetPathType, text_concat, text_parallel


class MetadataItem:
    def __init__(self, data: Mapping, segments: List[str]):
        self.data = data
        self.segments = segments

    def __hash__(self):
        return hash_anything((self.__class__, self.data, self.segments))

    def __eq__(self, other):
        if self is other:
            return True
        elif isinstance(other, MetadataItem):
            return self.data == other.data and self.segments == other.segments
        else:
            return False

    def __repr__(self):
        return f'<{self.__class__.__name__} data: {self.data!r}>'


def _metadata_repr(metadata: Mapping):
    with io.StringIO() as f:
        for key in sorted(metadata.keys()):
            print(f'{key}: {metadata[key]!r}', file=f)

        return f.getvalue()


class SyncTree:
    def __init__(self):
        self.metadata: dict = {}
        self.items: Dict[str, Union[SyncItem, SyncTree]] = {}

    def _add_item_with_segment(self, item: SyncItem, segments: List[str]):
        current = segments[0]
        if len(segments) > 1:  # still a path
            if current not in self.items or isinstance(self.items[current], SyncTree):
                if current not in self.items:
                    self.items[current] = SyncTree()
                self.items[current]._add_item_with_segment(item, segments[1:])
            elif isinstance(self.items[current], SyncItem):
                raise TypeError(f'Sync item ({"/".join(segments)}) is an item, '
                                f'unable to be set a child item ({item!r}) for it.')
            else:
                assert False, 'Unable to reach here #1, ' \
                              'please open an issue to let the author know.'  # pragma: no cover

        else:  # is a file
            if current not in self.items:
                self.items[current] = item
            elif isinstance(self.items[current], SyncItem):
                warnings.warn(f'Sync item ({self.items[current]!r}) at {"/".join(segments)} '
                              f'will be replaced by {item!r}.')
                self.items[current] = item
            elif isinstance(self.items[current], SyncTree):
                raise TypeError(f'Sync position ({"/".join(segments)}) is a folder, '
                                f'unable to be replace with item {item!r}.')
            else:
                assert False, 'Unable to reach here #2, ' \
                              'please open an issue to let the author know.'  # pragma: no cover

    def add_item(self, item: SyncItem):
        self._add_item_with_segment(item, item.segments)

    def _add_meta_item_with_segment(self, item: MetadataItem, segments: List[str]):
        if not segments:
            self.metadata.update(**item.data)
        else:
            current = segments[0]
            if current not in self.items:
                self.items[current] = SyncTree()

            if isinstance(self.items[current], SyncItem):
                if len(segments) == 1:
                    self.items[current].metadata.update(**item.data)
                else:
                    raise TypeError(f'Unable to set metadata {item.data!r} with position {"/".join(segments)} to '
                                    f'the child position of an item {"/".join(self.items[current].segments)}.')
            elif isinstance(self.items[current], SyncTree):
                self.items[current]._add_meta_item_with_segment(item, segments[1:])
            else:
                assert False, 'Unable to reach here #3, ' \
                              'please open an issue to let the author know.'  # pragma: no cover

    def add_meta_item(self, item: MetadataItem):
        self._add_meta_item_with_segment(item, item.segments)

    def _data_for_repr(self):
        def _recursion(tree, path):
            if isinstance(tree, SyncTree):
                label = text_concat(path, '[Metadata]', _metadata_repr(tree.metadata)) if tree.metadata else path
                return label, [
                    _recursion(tree.items[key], key)
                    for key in sorted(tree.items.keys())
                ]
            else:
                body = text_concat(repr(tree), '[Metadata]', _metadata_repr(tree.metadata)) \
                    if tree.metadata else repr(tree)
                return text_parallel(f'{path} --> ', body), []

        return _recursion(self, '<root>')

    def __repr__(self):
        return format_tree(self._data_for_repr(), itemgetter(0), itemgetter(1))

    def __hash__(self):
        return hash_anything((type(self), self.metadata, self.items))

    def __eq__(self, other):
        if self is other:
            return True
        elif isinstance(other, SyncTree):
            return self.items == other.items and self.metadata == other.metadata
        else:
            return False


SyncItemType = Union[SyncItem, MetadataItem]


class SyncResource:
    def __init__(self):
        self._registered_ops: Dict[str, Callable] = {}

    def grab(self) -> Iterable[Union[
        Tuple[str, Any, TargetPathType, Mapping],
        Tuple[str, Any, TargetPathType],
    ]]:
        raise NotImplementedError  # pragma: no cover

    def register_operation(self, name: str, func: Callable):
        if name in _PRESERVED_NAMES or name in _REGISTERED_SYNC_TYPES:
            raise KeyError(f'Operation name {name!r} has already been preserved or registered as a sync item.')
        elif name not in self._registered_ops:
            self._registered_ops[name] = func
        else:
            raise KeyError(f'Operation {name} has already been registered.')

    def iter_sync_items(self) -> Iterable[SyncItemType]:
        for tpl in self.grab():
            if isinstance(tpl, tuple):
                if len(tpl) == 3:
                    (type_, value, path), attached_data = tpl, {}
                elif len(tpl) == 4:
                    type_, value, path, attached_data = tpl
                else:
                    raise ValueError(f'Sync type data should be a tuple '
                                     f'with length 3 or 4, but {tpl!r} found.')

                segments = to_segments(path)
                if type_ in _PRESERVED_NAMES:
                    if type_ == 'metadata':
                        if attached_data:
                            warnings.warn(f'Attached data {attached_data!r} for position {tpl!r} '
                                          f'will be ignored when defining metadata.')
                        yield MetadataItem(value, segments)

                    else:
                        assert False, f'Undefined preserved operation - {type_!r}, ' \
                                      f'please notice the author about this.'  # pragma: no cover
                elif type_ in self._registered_ops:
                    self._registered_ops[type_](value, segments, attached_data)
                else:
                    yield create_sync_item(type_, value, attached_data, segments)

            else:
                raise TypeError(f'Invalid sync type data - {tpl!r}.')

    def sync_tree(self) -> SyncTree:
        tree = SyncTree()
        meta_items: List[MetadataItem] = []
        for item in self.iter_sync_items():
            if isinstance(item, MetadataItem):
                meta_items.append(item)
            else:
                tree.add_item(item)

        for item in meta_items:
            tree.add_meta_item(item)

        return tree



import os

import requests
from tqdm.auto import tqdm

from .session import get_requests_session, srequest


def download_file(url, filename, expected_size: int = None, desc=None, session=None, **kwargs):
    session = session or get_requests_session()
    response = srequest(session, 'GET', url, stream=True, allow_redirects=True, **kwargs)
    expected_size = expected_size or response.headers.get('Content-Length', None)
    expected_size = int(expected_size) if expected_size is not None else expected_size

    desc = desc or os.path.basename(filename)
    directory = os.path.dirname(filename)
    if directory:
        os.makedirs(directory, exist_ok=True)

    with open(filename, 'wb') as f:
        with tqdm(total=expected_size, unit='B', unit_scale=True, unit_divisor=1024, desc=desc) as pbar:
            for chunk in response.iter_content(chunk_size=1024):
                f.write(chunk)
                pbar.update(len(chunk))

    actual_size = os.path.getsize(filename)
    if expected_size is not None and actual_size != expected_size:
        os.remove(filename)
        raise requests.exceptions.HTTPError(f"Downloaded file is not of expected size, "
                                            f"{expected_size} expected but {actual_size} found.")

    return filename



import os
from typing import Iterable, Union, Tuple, Any, Mapping

from .resource import SyncResource
from ..utils import TargetPathType


class LocalDirectoryResource(SyncResource):
    def __init__(self, directory):
        SyncResource.__init__(self)
        self.directory = directory

    def grab(self) -> Iterable[Union[
        Tuple[str, Any, TargetPathType, Mapping],
        Tuple[str, Any, TargetPathType],
    ]]:
        for root, dirs, files in os.walk(self.directory):
            for file in files:
                path = os.path.abspath(os.path.join(root, file))
                relpath = os.path.relpath(path, self.directory)
                yield 'local', path, relpath



from collections.abc import Mapping, Sequence


def _object_hashable(obj):
    try:
        _ = hash(obj)
        return obj
    except TypeError:
        if isinstance(obj, Mapping):
            items = []
            for key in sorted(obj.keys(), key=repr):
                items.append((_object_hashable(key), _object_hashable(obj[key])))
            return type(obj), tuple(items)
        elif isinstance(obj, Sequence):
            items = [_object_hashable(item) for item in obj]
            return type(obj), tuple(items)
        else:
            return type(obj), id(obj)


def hash_anything(obj):
    return hash(_object_hashable(obj))



import time
from typing import Optional, Dict

import requests
from requests.adapters import HTTPAdapter, Retry
from requests.exceptions import RequestException

DEFAULT_TIMEOUT = 10  # seconds


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.timeout = DEFAULT_TIMEOUT
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


def get_requests_session(max_retries: int = 5, timeout: int = DEFAULT_TIMEOUT,
                         headers: Optional[Dict[str, str]] = None) -> requests.Session:
    session = requests.session()
    retries = Retry(
        total=max_retries, backoff_factor=1,
        status_forcelist=[413, 429, 500, 501, 502, 503, 504, 505, 506, 507, 509, 510, 511],
        allowed_methods=["HEAD", "GET", "POST", "PUT", "DELETE", "OPTIONS", "TRACE"],
    )
    adapter = TimeoutHTTPAdapter(max_retries=retries, timeout=timeout)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    session.headers.update({
        "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 "
                      "(KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
        **dict(headers or {}),
    })

    return session


def srequest(session: requests.Session, method, url, *, max_retries: int = 5,
             sleep_time: float = 5.0, raise_for_status: bool = True, **kwargs) -> requests.Response:
    resp = None
    for _ in range(max_retries):
        try:
            resp = session.request(method, url, **kwargs)
        except RequestException:
            time.sleep(sleep_time)
        else:
            break
    assert resp is not None
    if raise_for_status:
        resp.raise_for_status()

    return resp



import os

from hbutils.random import random_sha1_with_timestamp
from hbutils.system import TemporaryDirectory, copy


class FilePool:
    def __init__(self):
        self.tmpdir = TemporaryDirectory()
        self.count = 0

    def __del__(self):
        self.tmpdir.cleanup()
        self.count = 0

    def put_file(self, path) -> str:
        tmppath = os.path.join(self.tmpdir.name, random_sha1_with_timestamp())
        os.makedirs(tmppath)

        dst_filename = os.path.join(tmppath, os.path.basename(os.path.abspath(path)))
        copy(path, dst_filename)

        self.count += 1
        return dst_filename

    def cleanup(self):
        self.tmpdir.cleanup()
        self.tmpdir = TemporaryDirectory()
        self.count = 0

    def __len__(self):
        return self.count



import json
import os.path
from itertools import chain
from typing import List, Tuple

from hbutils.reflection import nested_with
from hbutils.string import plural_word
from hbutils.system.filesystem.tempfile import TemporaryDirectory
from tqdm import tqdm as _TqdmType
from tqdm.auto import tqdm

from ..resource import SyncResource, SyncTree, ResourceNotChange
from ..storage import BaseStorage
from ..utils import FilePool


def _count_trees(tree: SyncTree):
    tree_cnt, file_cnt = 1, 0
    for _, item in tree.items.items():
        if isinstance(item, SyncTree):
            l_tree_cnt, l_file_cnt = _count_trees(item)
            tree_cnt += l_tree_cnt
            file_cnt += l_file_cnt
        else:
            file_cnt += 1

    return tree_cnt, file_cnt


class SyncTask:
    def __init__(self, resource: SyncResource, storage: BaseStorage, meta_filename='.meta.json',
                 batch: int = 50):
        self.resource = resource
        self.storage = storage
        self.meta_filename = meta_filename

        # batch < 0, do not submit until the end of sync
        # batch == 0, submit immediately every time
        # batch > 0, submit changes when changes over `batch`
        self.batch = batch

    def _sync_tree(self, tree: SyncTree, segments: List[str], tqdms: Tuple[_TqdmType, _TqdmType],
                   preserved: Tuple[FilePool, List]):
        tree_tqdm, file_tqdm = tqdms
        tree_tqdm.set_description('/'.join(segments))
        items, folders = [], []
        for key in sorted(tree.items.keys()):
            value = tree.items[key]
            if isinstance(value, SyncTree):
                folders.append((key, value))
            else:
                items.append((key, value))

        meta_file_segments = [*segments, self.meta_filename]
        if self.storage.file_exists(meta_file_segments):
            old_metadata = json.loads(self.storage.read_text(meta_file_segments))
            old_files = {item['name']: item for item in old_metadata['files']}
            old_item_names = {item['name'] for item in chain(old_metadata['files'], old_metadata['folders'])}
        else:
            old_files = {}
            old_item_names = set()

        m_folders = []
        for key, folder in folders:
            self._sync_tree(folder, [*segments, key], tqdms, preserved)
            m_folders.append({'name': key, 'metadata': folder.metadata})

        m_files = []
        need_load_files = []
        for key, item in tqdm(items, desc=f"Mark for {'/'.join(segments)}"):
            old_file_data = old_files.get(key)
            if old_file_data and old_file_data['type'] == item.__type__:
                try:
                    mark = item.refresh_mark(old_file_data['mark'])
                except ResourceNotChange:
                    need_load, mark = False, old_file_data['mark']
                else:
                    need_load = True
            else:
                need_load = True
                mark = item.refresh_mark(None)

            if need_load:
                need_load_files.append((key, item))
            else:
                file_tqdm.update()
                file_tqdm.set_description(plural_word(file_tqdm.n, 'file'))

            m_files.append({
                'name': key,
                'type': item.__type__,
                'mark': mark,
                'metadata': item.metadata,
            })

        file_pool, preserved_changes = preserved
        with TemporaryDirectory() as td:
            local_metafile = os.path.join(td, self.meta_filename)
            with open(local_metafile, 'w', encoding='utf-8') as f:
                json.dump({
                    'path': '/'.join(segments),
                    'metadata': tree.metadata,
                    'files': m_files,
                    'folders': m_folders,
                }, f, indent=4, ensure_ascii=False)

            new_item_names = {item['name'] for item in chain(m_files, m_folders)}
            with nested_with(*[item.load_file() for _, item in need_load_files]) as file_paths:
                changes = [(local_metafile, [*segments, self.meta_filename])]  # .meta.json
                for local_file, (key, _) in zip(file_paths, need_load_files):  # items to add
                    changes.append((local_file, [*segments, key]))
                for key in sorted(old_item_names - new_item_names):  # items to delete
                    changes.append((None, [*segments, key]))

                if self.batch == 0:
                    self.storage.batch_change_files(changes)
                else:
                    for local_file, remote_segs in changes:
                        if local_file is not None:
                            preserved_changes.append((file_pool.put_file(local_file), remote_segs))
                        else:
                            preserved_changes.append((None, remote_segs))

                    if 0 < self.batch <= len(preserved_changes):
                        self.storage.batch_change_files(preserved_changes)
                        preserved_changes.clear()
                        file_pool.cleanup()

            file_tqdm.update(len(need_load_files))
            file_tqdm.set_description(plural_word(file_tqdm.n, 'file'))

        tree_tqdm.update()

    def sync(self):
        tree: SyncTree = self.resource.sync_tree()
        total_trees, total_files = _count_trees(tree)

        tree_tqdm = tqdm(total=total_trees)
        file_tqdm = tqdm(total=total_files)
        file_pool, preserved_changes = FilePool(), []
        self._sync_tree(tree, [], (tree_tqdm, file_tqdm), (file_pool, preserved_changes))
        if preserved_changes:
            self.storage.batch_change_files(preserved_changes)
            preserved_changes.clear()
            file_pool.cleanup()



import os
from dataclasses import dataclass
from typing import Union


def text_concat(*texts):
    lines = []
    for text in texts:
        lines.extend(text.splitlines(keepends=False))
    return os.linesep.join(lines)


class _TextBlock:
    def __init__(self, text: str, keepends: bool = False, pad: bool = True):
        self._lines = text.splitlines(keepends=keepends)
        self._max_width = max(map(len, self._lines))
        self._pad = pad

    def _pad_line(self, line):
        if self._pad:
            return line + ' ' * (self._max_width - len(line))
        else:
            return line

    def __getitem__(self, item: int):
        if isinstance(item, int):
            if item >= len(self._lines):
                line = ''
            else:
                line = self._lines[item]

            return self._pad_line(line)
        else:
            raise TypeError(f'Supported item type - {item!r}.')  # pragma: no cover

    def __len__(self):
        return len(self._lines)


class _CycleTextBlock(_TextBlock):
    def __getitem__(self, item: int):
        return _TextBlock.__getitem__(self, item % len(self))


@dataclass
class _CycleTextBlockProxy:
    text: str


def _to_text_block(x, **kwargs) -> _TextBlock:
    if isinstance(x, _CycleTextBlockProxy):
        return _CycleTextBlock(x.text, **kwargs)
    elif isinstance(x, str):
        return _TextBlock(x, **kwargs)
    else:
        raise TypeError(f'Unknown type of text block - {x!r}.')


def cycle(text):
    return _CycleTextBlockProxy(text)


def text_parallel(*texts: Union[str, _CycleTextBlockProxy]):
    if not texts:
        return ''

    blocks = [_to_text_block(x, pad=i < (len(texts) - 1)) for i, x in enumerate(texts)]
    native_blocks = [x for x in blocks if not isinstance(x, _CycleTextBlock)]
    max_lines = max(map(len, native_blocks))

    lines = []
    for i in range(max_lines):
        parts = [b[i] for b in blocks]
        lines.append(''.join(parts))

    return os.linesep.join(lines)



from .download import download_file
from .filepool import FilePool
from .hash import hash_anything
from .segments import to_segments, TargetPathType
from .session import get_requests_session, srequest
from .text import text_concat, cycle, text_parallel



import datetime
import os
import warnings
from functools import partial, lru_cache
from hashlib import sha256, sha1
from typing import List, Tuple, Optional, Union, Dict

from huggingface_hub import HfApi, hf_hub_url, CommitOperationAdd, CommitOperationDelete, configure_http_backend

from .base import BaseStorage
from ..utils import to_segments, srequest, get_requests_session

DEFAULT_TIMEOUT: int = 10


@lru_cache()
def _register_session_for_hf(max_retries: int = 5, timeout: int = DEFAULT_TIMEOUT,
                             headers: Optional[Dict[str, str]] = None):
    configure_http_backend(backend_factory=partial(get_requests_session, max_retries, timeout, headers))


def _single_resource_is_duplicated(local_filename: str, is_lfs: bool, oid: str, filesize: int,
                                   chunk_for_hash: int = 1 << 20) -> bool:
    if filesize != os.path.getsize(local_filename):
        return False

    if is_lfs:
        sha = sha256()
    else:
        sha = sha1()
        sha.update(f'blob {filesize}\0'.encode('utf-8'))
    with open(local_filename, 'rb') as f:
        # make sure the big files will not cause OOM
        while True:
            data = f.read(chunk_for_hash)
            if not data:
                break
            sha.update(data)

    return sha.hexdigest() == oid


def hf_local_upload_check(uploads: List[Tuple[Optional[str], str]],
                          repo_id: str, repo_type='dataset', revision='main',
                          chunk_for_hash: int = 1 << 20, session=None) -> List[Tuple[bool, str]]:
    """
    Overview:
        Check resource on huggingface repo and local.

    :param uploads: Tuples of uploads, the first item is the local file, second item is the file in repo. When \
        first item is None, it means delete this item in repo.
    :param repo_id: Repository id, the same as that in huggingface library.
    :param repo_type: Repository type, the same as that in huggingface library.
    :param revision: Revision of repository, the same as that in huggingface library.
    :param chunk_for_hash: Chunk size for hashing calculation.
    :param session: Session of requests, will be auto created when not given.
    :return: Uploads are necessary or not, in form of lists of boolean.
    """
    if not uploads:
        return []

    session = session or get_requests_session()
    files_in_repo = [f for _, f in uploads]
    resp = srequest(
        session,
        'POST', f"https://huggingface.co/api/{repo_type}s/{repo_id}/paths-info/{revision}",
        json={"paths": files_in_repo},
    )
    online_file_info = {tuple(to_segments(item['path'])): item for item in resp.json()}

    checks = []
    for f_in_local, f_in_repo in uploads:
        fs_in_repo = tuple(to_segments(f_in_repo))
        f_meta = online_file_info.get(fs_in_repo, None)
        if not f_meta:
            if f_in_local is not None:  # not exist in repo, need to upload
                checks.append((True, 'file'))
            else:  # not exist in repo, do not need to delete
                checks.append((False, None))
        else:
            if f_in_local is not None:  # going to upload
                if 'lfs' in f_meta:  # is a lfs file
                    is_lfs, oid, filesize = True, f_meta['lfs']['oid'], f_meta['lfs']['size']
                else:  # not lfs
                    is_lfs, oid, filesize = False, f_meta['oid'], f_meta['size']

                if f_meta['type'] != 'file':
                    raise FileExistsError(f'Path {f_meta["path"]!r} is a {f_meta["type"]} on huggingface, '
                                          f'unable to replace it with local file {f_in_local!r}.')
                _is_duplicated = _single_resource_is_duplicated(f_in_local, is_lfs, oid, filesize, chunk_for_hash)
                checks.append((not _is_duplicated, f_meta['type']))  # exist, need to upload if not the same
            else:  # going to delete
                checks.append((True, f_meta['type']))

    return checks


def _check_repo_type(repo_type):
    if repo_type in {'model', 'dataset', 'space'}:
        return repo_type
    else:
        raise ValueError(f'Invalid huggingface repository type - {repo_type!r}.')


class HuggingfaceStorage(BaseStorage):
    def __init__(self, repo: str, repo_type: str = 'dataset', revision: str = 'main',
                 hf_client: Optional[HfApi] = None, access_token: Optional[str] = None,
                 namespace: Union[List[str], str, None] = None):
        if hf_client and access_token:
            warnings.warn('Huggingface client provided, so access token will be ignored.', stacklevel=2)
        self.hf_client = hf_client or HfApi(token=access_token)
        self.repo = repo
        self.repo_type = _check_repo_type(repo_type)
        self.revision = revision
        self.namespace = to_segments(namespace or [])
        self.session = get_requests_session()

    def path_join(self, path, *segments):
        return '/'.join((*self.namespace, path, *segments))

    def _file_url(self, file: List[str]):
        return hf_hub_url(self.repo, self.path_join(*file), repo_type=self.repo_type, revision=self.revision)

    def file_exists(self, file: List[str]) -> bool:
        resp = srequest(self.session, 'HEAD', self._file_url(file), raise_for_status=False)
        if resp.ok:  # file is here
            return True
        elif resp.status_code == 404:  # file not found
            return False
        else:  # network error
            resp.raise_for_status()  # pragma: no cover

    def read_text(self, file: List[str], encoding: str = 'utf-8') -> str:
        return srequest(self.session, 'GET', self._file_url(file)).content.decode(encoding=encoding)

    def batch_change_files(self, changes: List[Tuple[Optional[str], List[str]]]):
        _register_session_for_hf()

        _map_changes = {}
        for local_filename, file_in_repo in changes:
            sg = tuple(file_in_repo)
            if sg in _map_changes:
                fip = self.path_join(*file_in_repo)
                if _map_changes[sg] is None:
                    warnings.warn(f'Deletion of resource {fip!r} is not necessary, will be ignored.')
                else:
                    warnings.warn(f'Uploading of local resource {_map_changes[sg]!r} to {fip!r} '
                                  f'is not necessary, will be ignored.')

            _map_changes[sg] = local_filename

        uploads = [
            (local_filename, self.path_join(*file_in_repo))
            for file_in_repo, local_filename in _map_changes.items()
        ]
        uploads_is_needed = hf_local_upload_check(uploads, self.repo, self.repo_type, self.revision,
                                                  session=self.session)

        operations, op_items, additions, deletions = [], [], 0, 0
        for (local_filename, fip), (need, objtype) in zip(uploads, uploads_is_needed):
            if need:
                if local_filename is None:
                    if objtype == 'directory':  # a / should be at the end of path when deleting a folder
                        fip = f'{fip}/'
                        is_folder = True
                    else:
                        is_folder = False
                    operations.append(CommitOperationDelete(path_in_repo=fip, is_folder=is_folder))
                    op_items.append(f'-{fip}')
                    deletions += 1
                else:
                    operations.append(CommitOperationAdd(path_in_repo=fip, path_or_fileobj=local_filename))
                    op_items.append(f'+{fip}')
                    additions += 1

        if operations:
            current_time = datetime.datetime.now().astimezone().strftime('%Y-%m-%d %H:%M:%S %Z')
            msg = ', '.join(sorted(op_items))
            commit_message = f"{msg}, on {current_time}"
            self.hf_client.create_commit(
                self.repo, operations,
                commit_message=commit_message,
                repo_type=self.repo_type,
                revision=self.revision,
            )
            return additions, deletions, commit_message
        else:
            return additions, deletions, None



import warnings
from typing import Tuple, Mapping, Any, Union, Iterable, Optional

from github import Github
from github.GitRelease import GitRelease
from tqdm.auto import tqdm

from .resource import TargetPathType
from .version import VersionBasedResource


def _to_int(v: Union[str, int]) -> Union[str, int]:
    try:
        return int(v)
    except (TypeError, ValueError):
        return v


class GithubReleaseResource(VersionBasedResource):
    def __init__(self, repo: str, *,
                 github_client: Optional[Github] = None,
                 access_token: Optional[str] = None,
                 add_version_attachment: bool = True):
        VersionBasedResource.__init__(self, add_version_attachment)
        self.repo = repo
        if github_client and access_token:
            warnings.warn('Github client provided, so access token will be ignored.', stacklevel=2)
        self.github_client = github_client or Github(access_token)

    def _tag_filter(self, tag):
        return tag

    def _filename_filter(self, tag, filename):
        _ = tag
        return filename

    def grab_for_items(self) -> Iterable[Union[
        Tuple[str, Any, TargetPathType, Mapping],
        Tuple[str, Any, TargetPathType],
    ]]:
        repo = self.github_client.get_repo(self.repo)
        yield 'metadata', {'source': repo.html_url}, ''
        repo_tqdm = tqdm(repo.get_releases())
        for release in repo_tqdm:
            release: GitRelease
            tag_name = self._tag_filter(release.tag_name)
            repo_tqdm.set_description(tag_name)
            if not tag_name:
                continue

            has_file = False
            for asset in release.get_assets():
                filename = self._filename_filter(tag_name, asset.name)
                if not filename:
                    continue

                download_url = asset.browser_download_url
                metadata = {'tag': release.tag_name, 'filename': asset.name}
                yield 'remote', download_url, f'{tag_name}/{filename}', metadata
                has_file = True

            if has_file:
                yield 'version', tag_name, tag_name
                release_metadata = {'version': release.tag_name, 'title': release.title, 'url': release.html_url}
                yield 'metadata', release_metadata, tag_name



from .github import GithubReleaseResource
from .item import SyncItem, TextOutputSyncItem, RemoteSyncItem, CustomSyncItem, ResourceNotChange, LocalFileSyncItem
from .local import LocalDirectoryResource
from .resource import SyncResource, SyncTree, MetadataItem
from .sourceforge import SourceForgeFilesResource
from .version import VersionBasedResource



import re
import warnings
from typing import List, Iterable, Union, Mapping, Tuple, Any

from .resource import SyncResource
from ..utils import TargetPathType


def _to_int(v: Union[str, int]) -> Union[str, int]:
    try:
        return int(v)
    except (TypeError, ValueError):
        return v


class VersionBasedResource(SyncResource):
    def __init__(self, add_version_attachment: bool = True):
        SyncResource.__init__(self)
        self._versions = []
        self.register_operation('version', self._add_version)
        self.add_version_attachment = add_version_attachment

    def _add_version(self, version, segments: List[str], attached_data):
        _ = segments, attached_data
        self._versions.append(version)

    __version_pattern__ = r'^([a-zA-Z]+(\.|-)?)?(?P<version>[\d.]+)$'
    __version_file_prefix__ = 'LATEST_RELEASE'

    def _version_to_tuple(self, version):
        matching = re.fullmatch(self.__version_pattern__, version)
        if matching:
            version_text = matching.group('version')
            return tuple(map(_to_int, version_text.split('.')))
        else:
            raise ValueError(f'Invalid version for release - {version!r}.')

    def grab(self) -> Iterable[Union[
        Tuple[str, Any, TargetPathType, Mapping],
        Tuple[str, Any, TargetPathType],
    ]]:
        self._versions.clear()
        yield from self.grab_for_items()

        if self.add_version_attachment and self._versions:
            version_map = {}
            for version in self._versions:
                try:
                    _tuple = self._version_to_tuple(version)
                except ValueError:
                    warnings.warn(f'Version {version!r} does not match the regular expression, '
                                  f'so it will be ignored in version indexing.')
                    continue

                for i in range(len(_tuple) + 1):
                    _part_tuple = _tuple[:i]
                    if _part_tuple in version_map:
                        _exist_tuple, _exist_version = version_map[_part_tuple]
                        if _tuple > _exist_tuple:
                            version_map[_part_tuple] = (_tuple, version)
                    else:
                        version_map[_part_tuple] = (_tuple, version)

            for tuple_, (_, version) in version_map.items():
                if tuple_:
                    schema_file = f'{self.__version_file_prefix__}_{".".join(map(str, tuple_))}'
                else:
                    schema_file = self.__version_file_prefix__
                yield 'text', version, schema_file

    def grab_for_items(self) -> Iterable[Union[
        Tuple[str, Any, TargetPathType, Mapping],
        Tuple[str, Any, TargetPathType],
    ]]:
        raise NotImplementedError  # pragma: no cover



import abc
import os.path
import time
from contextlib import contextmanager
from email.utils import parsedate_to_datetime
from typing import List, ContextManager, Optional, Type, Any, Dict

import requests
from hbutils.string import truncate
from hbutils.system.filesystem.tempfile import TemporaryDirectory
from hbutils.system.network import urlsplit

from ..utils import download_file, srequest, get_requests_session, hash_anything


class ResourceNotChange(Exception):
    pass


class SyncItem(metaclass=abc.ABCMeta):
    __type__: str = None

    def __init__(self, value, metadata: dict, segments: List[str]):
        self._value = value
        self.metadata = metadata
        self.segments = segments

    def load_file(self) -> ContextManager[str]:
        raise NotImplementedError  # pragma: no cover

    def refresh_mark(self, mark: Optional[Dict[str, Any]]):
        return mark or {}

    def __hash__(self):
        return hash_anything((type(self), self._value, self.metadata, self.segments))

    def __eq__(self, other):
        if self is other:
            return True
        elif isinstance(self, type(other)) and isinstance(other, type(self)):
            return (self._value, self.metadata, self.segments) == \
                (other._value, other.metadata, other.segments)
        else:
            return False


class RemoteSyncItem(SyncItem):
    __type__ = 'remote'
    __headers__ = {}
    __request_kwargs__ = {}

    def __init__(self, url, metadata, segments: List[str]):
        SyncItem.__init__(self, url, metadata, segments)
        self.url = url
        self._session = None

    def get_new_session(self):
        return get_requests_session(headers=self.__headers__)

    def _get_session(self) -> requests.Session:
        if self._session is None:
            self._session = self.get_new_session()

        return self._session

    def _file_process(self, filename):
        pass

    @contextmanager
    def load_file(self) -> ContextManager[str]:
        with TemporaryDirectory() as td:
            filename = os.path.join(td, urlsplit(self.url).filename or 'unnamed_file')
            download_file(self.url, filename, session=self._get_session(), **self.__request_kwargs__)
            self._file_process(filename)
            yield filename

    def refresh_mark(self, mark: Optional[Dict[str, Any]]):
        mark = dict(mark or {})
        url = mark.get('url')
        if url == self.url:  # url not changed
            expires = mark.get('expires')
            if expires is not None and time.time() < expires:
                raise ResourceNotChange

            etag = mark.get('etag')
            headers = {'If-None-Match': etag} if etag else {}
        else:
            headers = {}

        kwargs = self.__request_kwargs__.copy()
        if 'headers' in kwargs:
            headers.update(kwargs['headers'])
            kwargs.pop('headers')

        resp = srequest(
            self._get_session(), 'HEAD', self.url,
            allow_redirects=True, headers=headers,
            **kwargs
        )
        if resp.status_code == 304:
            raise ResourceNotChange

        headers = resp.headers
        etag = headers.get('ETag')
        expires = headers.get('Expires')
        expires = parsedate_to_datetime(expires).timestamp() if expires else None
        content_length = headers.get('Content-Length')
        content_length = int(content_length) if content_length is not None else None
        content_type = headers.get('Content-Type')

        return {
            'url': self.url,
            'etag': etag,
            'expires': expires,
            'content_length': content_length,
            'content_type': content_type,
        }

    def __repr__(self):
        return f'<{self.__class__.__name__} url: {self.url!r}>'


class CustomSyncItem(SyncItem):
    __type__ = 'custom'

    def __init__(self, gene, metadata, segments):
        SyncItem.__init__(self, gene, metadata, segments)
        self.gene = gene

    @contextmanager
    def load_file(self) -> ContextManager[str]:
        with self.gene() as f:
            yield f

    def __repr__(self):
        return f'<{self.__class__.__name__} gene: {self.gene!r}>'


class TextOutputSyncItem(SyncItem):
    __type__ = 'text'

    def __init__(self, content, metadata, segments):
        SyncItem.__init__(self, content, metadata, segments)
        self.content = content

    @contextmanager
    def load_file(self) -> ContextManager[str]:
        with TemporaryDirectory() as td:
            filename = os.path.join(td, 'file')
            with open(filename, 'w') as f:
                f.write(self.content)

            yield filename

    def __repr__(self):
        return f'<{self.__class__.__name__} content: {truncate(self.content, tail_length=15, show_length=True)!r}>'


class LocalFileSyncItem(SyncItem):
    __type__ = 'local'

    def __init__(self, filename, metadata, segments):
        SyncItem.__init__(self, filename, metadata, segments)
        self.filename = filename

    @contextmanager
    def load_file(self) -> ContextManager[str]:
        yield self.filename


_PRESERVED_NAMES = {'metadata'}
_REGISTERED_SYNC_TYPES: Dict[str, Type[SyncItem]] = {}


def register_sync_type(clazz: Type[SyncItem]):
    type_ = clazz.__type__
    if type_ is None:
        raise TypeError(f'Sync item class {clazz!r} should have __type__ instead of None.')
    elif type_ in _PRESERVED_NAMES:
        raise KeyError(f'Type {type_!r} is preserved, please use another one.')
    elif type_ in _REGISTERED_SYNC_TYPES:
        raise KeyError(f'Sync item type {type_!r} already exist.')
    else:
        _REGISTERED_SYNC_TYPES[type_] = clazz


register_sync_type(RemoteSyncItem)
register_sync_type(CustomSyncItem)
register_sync_type(TextOutputSyncItem)
register_sync_type(LocalFileSyncItem)


def create_sync_item(type_: str, value: Any, metadata: dict, segments: List[str]) -> SyncItem:
    if type_ in _REGISTERED_SYNC_TYPES:
        clazz = _REGISTERED_SYNC_TYPES[type_]
        return clazz(value, metadata, segments)
    else:
        raise KeyError(f'Unknown sync item type - {type_!r}.')



import os
import re
from typing import List, Union

TargetPathType = Union[str, List[str]]

_FORBIDDEN_CHARS = {'<', '>', ':', '\"', '/', '\\', '|', '?', '*', *(chr(ich) for ich in range(1, 32))}
_FORBIDDEN_SEGMENTS = [
    'CON', 'PRN', 'AUX', 'NUL',
    'COM1', 'COM2', 'COM3', 'COM4', 'COM5', 'COM6', 'COM7', 'COM8', 'COM9',
    'LPT1', 'LPT2', 'LPT3', 'LPT4', 'LPT5', 'LPT6', 'LPT7', 'LPT8', 'LPT9',
]


def to_segments(path: TargetPathType) -> List[str]:
    if isinstance(path, (list, tuple)):
        segments = map(str, path)
    else:
        segments = re.split(r'[\\/]+', path)
    segments = list(filter(bool, segments))

    for i, segment in enumerate(segments):
        for ch in segment:
            if ch in _FORBIDDEN_CHARS:
                raise ValueError(f'Segment #{i} contains invalid character - {segment!r}.')

        for fseg in _FORBIDDEN_SEGMENTS:
            if fseg.lower() == segment.lower():
                raise ValueError(f'Segment #{i} is preserved - {segment!r}.')

    normed_path = os.path.relpath(os.path.normpath(os.path.join(os.path.sep, *segments)), start=os.path.sep)
    segments = normed_path.split(os.path.sep)
    return [] if segments == ['.'] else segments



import re
from cgi import parse_header
from typing import Iterable, Tuple, Optional, List
from typing import Union, Any, Mapping
from urllib import parse as urlparse
from urllib.parse import urljoin, quote

from pyquery import PyQuery as pq
from tqdm.auto import tqdm

from .item import RemoteSyncItem, register_sync_type
from .version import VersionBasedResource
from ..utils import TargetPathType, srequest, get_requests_session, to_segments


def _iterdir_on_sourceforge(url, segments, session=None) -> Iterable[Tuple[str, str, str]]:
    session = session or get_requests_session()
    resp = srequest(session, 'GET', url)

    list_tqdm = tqdm(list(pq(resp.text)('#files_list tbody > tr').items()))
    for row in list_tqdm:
        th = row('th:nth-child(1)')
        name = th('.name').text().strip()
        list_tqdm.set_description('/'.join([*segments, name]))
        download_page_url = urljoin(resp.url, th('a').attr('href'))
        type_, = re.findall(r'(download|enter)', th('a').attr('title'))
        if type_ == 'download':
            r = srequest(session, 'GET', download_page_url)
            raw_download_url = None
            for item in parse_header(pq(r.text)('noscript meta').attr('content')):
                if isinstance(item, dict) and 'url' in item:
                    raw_download_url = item['url']
                    break

            assert raw_download_url, f'Raw download url not found on {download_page_url!r}.'
            download_url = urlparse.urlunsplit(urlparse.urlsplit(raw_download_url)._replace(query=''))
            yield 'file', name, download_url
        else:
            yield 'directory', name, download_page_url


class WgetRemoteItem(RemoteSyncItem):
    __type__ = 'wget'
    __headers__ = {'User-Agent': 'Wget/1.20.3 (linux-gnu)'}


register_sync_type(WgetRemoteItem)


class SourceForgeFilesResource(VersionBasedResource):
    def __init__(self, project_name, subdir='', add_version_attachment: bool = True):
        VersionBasedResource.__init__(self, add_version_attachment)
        self.project_name = project_name
        self.subdir = to_segments(subdir)
        self.root_url = f"https://sourceforge.net/projects/" \
                        f"{quote(project_name)}/files/{'/'.join(map(quote, self.subdir))}"

    def _process_segments(self, type_, segments) -> Optional[List[str]]:
        _ = type_, segments
        return segments

    def _get_version(self, type_, segments) -> Optional[str]:
        _ = type_, segments
        return None

    def _walk_on_sourceforge(self, url, segments, session=None):
        session = session or get_requests_session()
        for type_, name, download_url in _iterdir_on_sourceforge(url, segments, session):
            current_segments = [*segments, name]
            _to_segments = self._process_segments(type_, current_segments)
            if _to_segments is not None:
                yield type_, _to_segments, download_url
                if type_ == 'directory':
                    yield from self._walk_on_sourceforge(download_url, current_segments, session)

    def grab_for_items(self) -> Iterable[Union[
        Tuple[str, Any, TargetPathType, Mapping],
        Tuple[str, Any, TargetPathType],
    ]]:
        yield 'metadata', {'source': self.root_url}, ''
        session = get_requests_session()
        for type_, segments, download_url in self._walk_on_sourceforge(self.root_url, [], session):
            if type_ == 'file':
                yield 'wget', download_url, segments
            else:  # directory
                yield 'metadata', {'page_url': download_url}, segments

            version = self._get_version(type_, segments)
            if version is not None:
                yield 'version', version, segments



import os.path
import pathlib
from contextlib import contextmanager
from typing import List, Optional, Tuple, Union

from hbutils.system import copy, remove
from hbutils.system.filesystem.tempfile import TemporaryDirectory

from .base import BaseStorage
from ..utils import to_segments


class LocalStorage(BaseStorage):
    def __init__(self, root_directory, namespace: Union[List[str], str, None] = None):
        self.root_directory = root_directory
        self.namespace = to_segments(namespace or [])

    def path_join(self, path, *segments):
        return os.path.join(self.root_directory, *self.namespace, path, *segments)

    def file_exists(self, file: List[str]) -> bool:
        file = self.path_join(*file)
        return os.path.exists(file) and os.path.isfile(file)

    def read_text(self, file: List[str], encoding: str = 'utf-8') -> str:
        file = self.path_join(*file)
        return pathlib.Path(file).read_text(encoding=encoding)

    @contextmanager
    def recover_state_when_failed(self, state: List[List[str]]):
        with TemporaryDirectory() as td:
            # record all the files
            records = []
            for i, file_in_storage in enumerate(state):
                file_in_storage = self.path_join(*file_in_storage)
                if os.path.exists(file_in_storage):
                    file_in_temp = os.path.join(td, f'{i}', os.path.basename(file_in_storage))
                    os.makedirs(os.path.join(td, f'{i}'), exist_ok=True)
                    copy(file_in_storage, file_in_temp)
                    records.append(file_in_temp)
                else:
                    records.append(None)

            try:
                yield
            except:
                # recover all the files
                for file_in_temp, file_in_storage in zip(records, state):
                    file_in_storage = self.path_join(*file_in_storage)
                    if file_in_temp is None:
                        if os.path.exists(file_in_storage):
                            remove(file_in_storage)
                    else:
                        directory = os.path.dirname(file_in_storage)
                        if directory:
                            os.makedirs(directory, exist_ok=True)
                        copy(file_in_temp, file_in_storage)

                raise

    def batch_change_files(self, changes: List[Tuple[Optional[str], List[str]]]):
        states = [file_in_storage for _, file_in_storage in changes]
        with self.recover_state_when_failed(states):
            for local_file, file_in_storage in changes:
                file_in_storage = self.path_join(*file_in_storage)
                if local_file is None:
                    if os.path.exists(file_in_storage):
                        remove(file_in_storage)
                else:
                    directory = os.path.dirname(file_in_storage)
                    if directory:
                        os.makedirs(directory, exist_ok=True)
                    copy(local_file, file_in_storage)
